// Copyright (C) 1996 DIMACS Center, Rutgers, The State University of New Jersey
// Author(s): Jonathan Berry

// This software is copyrighted by the DIMACS Center at Rutgers, The State
// University of New Jersey.  IT IS PROVIDED AS IS, AND THE AUTHORS, DIMACS, AND
// RUTGERS, THE STATE UNIVERSITY OF NEW JERSEY  DISCLAIM
// ALL LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
// DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
// DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

// THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
// IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
// NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
// MODIFICATIONS.

// The authors hereby grant permission to use, copy, modify, distribute,
// and license this software and its documentation for any purpose, provided
// that existing copyright notices are retained in all copies and that this
// notice is included verbatim in any distributions. No written agreement,
// license, or royalty fee is required for any of the authorized uses.
// Modifications to this software may be copyrighted by their authors
// and need not follow the licensing terms described here, provided that
// the new terms are clearly indicated on the first page of each file where
// they apply.

// Last File Update: 31-Jul-1996
// 

#ifndef Container_h
#define Container_h

#include <fstream.h>
#include <LINK/basic/general.h>
#include <LINK/basic/Iterator.h>
#include <LINK/basic/ElementOps.h> // if DEFINE_TEMPLATES, methods defs are
				   // needlessly included.  Rely on optimizer.
				   // alternative would be new constant:
				   // DEFINE_E_TEMPLATES.

//template <class Item> class Collection;
#include<iostream.h>

template <class Item> class Container;
template <class Item> ostream& operator<<(ostream& os, 
					  const Container<Item>& a);
class ContainerNode {
public:
    virtual void*	info() const { return (void*) 0;}
};

template <class Item> 
class Container {
protected:
    //Item		empty_data; // not good - return static item instead
public:
    virtual ~Container() {}
    virtual Item	info(const ContainerNode* cn) const;
    virtual int		size() const = 0;
    virtual Bool	emptyQ() const = 0;
    virtual Bool	fullQ() const = 0;
    virtual Bool	memberQ(const Item& e) const = 0;
    virtual Bool	sortedQ() const = 0;
    virtual Item	max() const;
    virtual Item	min() const;
    virtual DataType	type() const = 0;

    virtual Item	get() = 0;
    virtual void        remove(const Item e) = 0;
    virtual void	clear() = 0;
    virtual int         iterate(Iterator<Item>& i, Item& e) const = 0; 
    virtual ostream&    display(ostream& os) const = 0;

    virtual Bool	operator==(const Container<Item>& m) const;
    virtual Bool	operator<=(const Container<Item>& m) const;
    virtual Bool	operator<(const Container<Item>& m) const;
    virtual Bool	operator>=(const Container<Item>& m) const;
    virtual Bool	operator>(const Container<Item>& m) const;
    virtual Bool	operator!=(const Container<Item>& m) const;

    virtual Bool	operator==(const Collection<Item>& m) const;
    virtual Bool	operator<=(const Collection<Item>& m) const;
    virtual Bool	operator<(const Collection<Item>& m) const;
    virtual Bool	operator>=(const Collection<Item>& m) const;
    virtual Bool	operator>(const Collection<Item>& m) const;
    virtual Bool	operator!=(const Collection<Item>& m) const;
    

    friend ostream& operator<< <>(ostream& os, const Container<Item>& a);
              
    virtual int		getRefCount()	{ return ref_count; }
    virtual void	initRefCount()	{ ref_count = 1;    }
    virtual void	incrRefCount()	{ ref_count++;	    }
    virtual void	decrRefCount()  { ref_count--;      }
protected:
    int     ref_count;
};


template <class Item>
class SimpleContainer : public Container<Item> {
public:
    virtual ContainerNode*	insert(const Item e) = 0;
    virtual ContainerNode*	append(const Item e) = 0;
    void	 		concatenate(const Container<Item>& c);
private:
    int				getRefCount() 
					{return Container<Item>::getRefCount();}
    void			initRefCount()
					{ Container<Item>::initRefCount(); }
    void			incrRefCount()
					{ Container<Item>::incrRefCount(); }
    void			decrRefCount()
					{ Container<Item>::decrRefCount(); }
};


template <class Key, class Item>
class Dictionary : public Container<Item> {
public:
    virtual ContainerNode*	insert(const Key k, const Item i) = 0;
private:
    int				getRefCount() 
					{return Container<Item>::getRefCount();}
    void			initRefCount()
					{ Container<Item>::initRefCount(); }
    void			incrRefCount()
					{ Container<Item>::incrRefCount(); }
    void			decrRefCount()
					{ Container<Item>::decrRefCount(); }
};

//#ifdef DEFINE_Container_TEMPLATE
#include <LINK/basic/Container.cpp>
//#endif

#endif
